"""Module that contains the command line app."""
import builtins
import inspect
import logging
import warnings
from os import path
from os import remove

import click
import matplotlib.pyplot as plt
import numpy as np
import yaml

from . import _cfg
from . import cache_tools
from . import global_params
from . import plotting
from . import wrapper as lib


def _get_config(config=None):
    if config is None:
        config = path.expanduser(path.join("~", ".21cmfast", "runconfig_example.yml"))

    with open(config, "r") as f:
        cfg = yaml.load(f, Loader=yaml.FullLoader)

    return cfg


def _ctx_to_dct(args):
    dct = {}
    j = 0
    while j < len(args):
        arg = args[j]
        if "=" in arg:
            a = arg.split("=")
            dct[a[0].replace("--", "")] = a[-1]
            j += 1
        else:
            dct[arg.replace("--", "")] = args[j + 1]
            j += 2

    return dct


def _update(obj, ctx):
    # Try to use the extra arguments as an override of config.
    kk = list(ctx.keys())
    for k in kk:
        # noinspection PyProtectedMember
        if hasattr(obj, k):
            try:
                val = getattr(obj, "_" + k)
                setattr(obj, "_" + k, type(val)(ctx[k]))
                ctx.pop(k)
            except (AttributeError, TypeError):
                try:
                    val = getattr(obj, k)
                    setattr(obj, k, type(val)(ctx[k]))
                    ctx.pop(k)
                except AttributeError:
                    pass


def _override(ctx, *param_dicts):
    # Try to use the extra arguments as an override of config.

    if ctx.args:
        ctx = _ctx_to_dct(ctx.args)
        for p in param_dicts:
            _update(p, ctx)

        # Also update globals, always.
        _update(global_params, ctx)
        if ctx:
            warnings.warn("The following arguments were not able to be set: %s" % ctx)


main = click.Group()


@main.command(
    context_settings={  # Doing this allows arbitrary options to override config
        "ignore_unknown_options": True,
        "allow_extra_args": True,
    }
)
@click.option(
    "--config",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Path to the configuration file (default ~/.21cmfast/runconfig_single.yml)",
)
@click.option(
    "--regen/--no-regen",
    default=False,
    help="Whether to force regeneration of init/perturb files if they already exist.",
)
@click.option(
    "--direc",
    type=click.Path(exists=True, dir_okay=True),
    default=None,
    help="directory to write data and plots to -- must exist.",
)
@click.option(
    "--seed",
    type=int,
    default=None,
    help="specify a random seed for the initial conditions",
)
@click.pass_context
def init(ctx, config, regen, direc, seed):
    """Run a single iteration of 21cmFAST init, saving results to file.

    Parameters
    ----------
    ctx :
        A parameter from the parent CLI function to be able to override config.
    config : str
        Path to the configuration file.
    regen : bool
        Whether to regenerate all data, even if found in cache.
    direc : str
        Where to search for cached items.
    seed : int
        Random seed used to generate data.
    """
    cfg = _get_config(config)

    # Set user/cosmo params from config.
    user_params = lib.UserParams(**cfg.get("user_params", {}))
    cosmo_params = lib.CosmoParams(**cfg.get("cosmo_params", {}))

    _override(ctx, user_params, cosmo_params)

    lib.initial_conditions(
        user_params=user_params,
        cosmo_params=cosmo_params,
        regenerate=regen,
        write=True,
        direc=direc,
        random_seed=seed,
    )


@main.command(
    context_settings={  # Doing this allows arbitrary options to override config
        "ignore_unknown_options": True,
        "allow_extra_args": True,
    }
)
@click.argument("redshift", type=float)
@click.option(
    "--config",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Path to the configuration file (default ~/.21cmfast/runconfig_single.yml)",
)
@click.option(
    "--regen/--no-regen",
    default=False,
    help="Whether to force regeneration of init/perturb files if they already exist.",
)
@click.option(
    "--direc",
    type=click.Path(exists=True, dir_okay=True),
    default=None,
    help="directory to write data and plots to -- must exist.",
)
@click.option(
    "--seed",
    type=int,
    default=None,
    help="specify a random seed for the initial conditions",
)
@click.pass_context
def perturb(ctx, redshift, config, regen, direc, seed):
    """Run 21cmFAST perturb_field at the specified redshift, saving results to file.

    Parameters
    ----------
    ctx :
        A parameter from the parent CLI function to be able to override config.
    redshift : float
        Redshift at which to generate perturbed field.
    config : str
        Path to the configuration file.
    regen : bool
        Whether to regenerate all data, even if found in cache.
    direc : str
        Where to search for cached items.
    seed : int
        Random seed used to generate data.
    """
    cfg = _get_config(config)

    # Set user/cosmo params from config.
    user_params = lib.UserParams(**cfg.get("user_params", {}))
    cosmo_params = lib.CosmoParams(**cfg.get("cosmo_params", {}))

    _override(ctx, user_params, cosmo_params)

    lib.perturb_field(
        redshift=redshift,
        user_params=user_params,
        cosmo_params=cosmo_params,
        regenerate=regen,
        write=True,
        direc=direc,
        random_seed=seed,
    )


@main.command(
    context_settings={  # Doing this allows arbitrary options to override config
        "ignore_unknown_options": True,
        "allow_extra_args": True,
    }
)
@click.argument("redshift", type=float)
@click.option(
    "-p",
    "--prev_z",
    type=float,
    default=None,
    help="Previous redshift (the spin temperature data must already exist for this redshift)",
)
@click.option(
    "--config",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Path to the configuration file (default ~/.21cmfast/runconfig_single.yml)",
)
@click.option(
    "--regen/--no-regen",
    default=False,
    help="Whether to force regeneration of init/perturb files if they already exist.",
)
@click.option(
    "--direc",
    type=click.Path(exists=True, dir_okay=True),
    default=None,
    help="directory to write data and plots to -- must exist.",
)
@click.option(
    "--seed",
    type=int,
    default=None,
    help="specify a random seed for the initial conditions",
)
@click.pass_context
def spin(ctx, redshift, prev_z, config, regen, direc, seed):
    """Run spin_temperature at the specified redshift, saving results to file.

    Parameters
    ----------
    ctx :
        A parameter from the parent CLI function to be able to override config.
    redshift : float
        The redshift to generate the field at.
    prev_z : float
        The redshift of a previous box from which to evolve to the current one.
    config : str
        Path to the configuration file.
    regen : bool
        Whether to regenerate all data, even if found in cache.
    direc : str
        Where to search for cached items.
    seed : int
        Random seed used to generate data.
    """
    cfg = _get_config(config)

    # Set user/cosmo params from config.
    user_params = lib.UserParams(**cfg.get("user_params", {}))
    cosmo_params = lib.CosmoParams(**cfg.get("cosmo_params", {}))
    flag_options = lib.FlagOptions(**cfg.get("flag_options", {}))
    astro_params = lib.AstroParams(
        **cfg.get("astro_params", {}), INHOMO_RECO=flag_options.INHOMO_RECO
    )

    _override(ctx, user_params, cosmo_params, astro_params, flag_options)

    lib.spin_temperature(
        redshift=redshift,
        astro_params=astro_params,
        flag_options=flag_options,
        previous_spin_temp=prev_z,
        user_params=user_params,
        cosmo_params=cosmo_params,
        regenerate=regen,
        write=True,
        direc=direc,
        random_seed=seed,
    )


@main.command(
    context_settings={  # Doing this allows arbitrary options to override config
        "ignore_unknown_options": True,
        "allow_extra_args": True,
    }
)
@click.argument("redshift", type=float)
@click.option(
    "-p",
    "--prev_z",
    type=float,
    default=None,
    help="Previous redshift (the ionized box data must already exist for this redshift)",
)
@click.option(
    "--config",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Path to the configuration file (default ~/.21cmfast/runconfig_single.yml)",
)
@click.option(
    "--regen/--no-regen",
    default=False,
    help="Whether to force regeneration of init/perturb files if they already exist.",
)
@click.option(
    "--direc",
    type=click.Path(exists=True, dir_okay=True),
    default=None,
    help="directory to write data and plots to -- must exist.",
)
@click.option(
    "--seed",
    type=int,
    default=None,
    help="specify a random seed for the initial conditions",
)
@click.pass_context
def ionize(ctx, redshift, prev_z, config, regen, direc, seed):
    """Run 21cmFAST ionize_box at the specified redshift, saving results to file.

    Parameters
    ----------
    ctx :
        A parameter from the parent CLI function to be able to override config.
    redshift : float
        The redshift to generate the field at.
    prev_z : float
        The redshift of a previous box from which to evolve to the current one.
    config : str
        Path to the configuration file.
    regen : bool
        Whether to regenerate all data, even if found in cache.
    direc : str
        Where to search for cached items.
    seed : int
        Random seed used to generate data.
    """
    cfg = _get_config(config)

    # Set user/cosmo params from config.
    user_params = lib.UserParams(**cfg.get("user_params", {}))
    cosmo_params = lib.CosmoParams(**cfg.get("cosmo_params", {}))
    flag_options = lib.FlagOptions(**cfg.get("flag_options", {}))
    astro_params = lib.AstroParams(
        **cfg.get("astro_params", {}), INHOMO_RECO=flag_options.INHOMO_RECO
    )

    _override(ctx, user_params, cosmo_params, astro_params, flag_options)

    lib.ionize_box(
        redshift=redshift,
        astro_params=astro_params,
        flag_options=flag_options,
        previous_ionize_box=prev_z,
        user_params=user_params,
        cosmo_params=cosmo_params,
        regenerate=regen,
        write=True,
        direc=direc,
        random_seed=seed,
    )


@main.command(
    context_settings={  # Doing this allows arbitrary options to override config
        "ignore_unknown_options": True,
        "allow_extra_args": True,
    }
)
@click.argument("redshift", type=str)
@click.option(
    "--config",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Path to the configuration file (default ~/.21cmfast/runconfig_single.yml)",
)
@click.option(
    "--regen/--no-regen",
    default=False,
    help="Whether to force regeneration of init/perturb files if they already exist.",
)
@click.option(
    "--direc",
    type=click.Path(exists=True, dir_okay=True),
    default=None,
    help="directory to write data and plots to -- must exist.",
)
@click.option(
    "--seed",
    type=int,
    default=None,
    help="specify a random seed for the initial conditions",
)
@click.pass_context
def coeval(ctx, redshift, config, regen, direc, seed):
    """Efficiently generate coeval cubes at a given redshift.

    Parameters
    ----------
    ctx :
        A parameter from the parent CLI function to be able to override config.
    redshift : float
        The redshift to generate the field at.
    config : str
        Path to the configuration file.
    regen : bool
        Whether to regenerate all data, even if found in cache.
    direc : str
        Where to search for cached items.
    seed : int
        Random seed used to generate data.
    """
    try:
        redshift = [float(z.strip()) for z in redshift.split(",")]
    except TypeError:
        raise TypeError("redshift argument must be comma-separated list of values.")

    cfg = _get_config(config)

    # Set user/cosmo params from config.
    user_params = lib.UserParams(**cfg.get("user_params", {}))
    cosmo_params = lib.CosmoParams(**cfg.get("cosmo_params", {}))
    flag_options = lib.FlagOptions(**cfg.get("flag_options", {}))
    astro_params = lib.AstroParams(
        **cfg.get("astro_params", {}), INHOMO_RECO=flag_options.INHOMO_RECO
    )

    _override(ctx, user_params, cosmo_params, astro_params, flag_options)

    lib.run_coeval(
        redshift=redshift,
        astro_params=astro_params,
        flag_options=flag_options,
        user_params=user_params,
        cosmo_params=cosmo_params,
        regenerate=regen,
        write=True,
        direc=direc,
        random_seed=seed,
    )


@main.command(
    context_settings={  # Doing this allows arbitrary options to override config
        "ignore_unknown_options": True,
        "allow_extra_args": True,
    }
)
@click.argument("redshift", type=float)
@click.option(
    "--config",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Path to the configuration file (default ~/.21cmfast/runconfig_single.yml)",
)
@click.option(
    "--regen/--no-regen",
    default=False,
    help="Whether to force regeneration of init/perturb files if they already exist.",
)
@click.option(
    "--direc",
    type=click.Path(exists=True, dir_okay=True),
    default=None,
    help="directory to write data and plots to -- must exist.",
)
@click.option(
    "-X",
    "--max-z",
    type=float,
    default=None,
    help="maximum redshift of the stored lightcone data",
)
@click.option(
    "--seed",
    type=int,
    default=None,
    help="specify a random seed for the initial conditions",
)
@click.pass_context
def lightcone(ctx, redshift, config, regen, direc, max_z, seed):
    """Efficiently generate coeval cubes at a given redshift.

    Parameters
    ----------
    ctx :
        A parameter from the parent CLI function to be able to override config.
    redshift : float
        The redshift to generate the field at.
    config : str
        Path to the configuration file.
    regen : bool
        Whether to regenerate all data, even if found in cache.
    direc : str
        Where to search for cached items.
    max_z : float
        Maximum redshift to include in the produced lightcone.
    seed : int
        Random seed used to generate data.
    """
    cfg = _get_config(config)

    # Set user/cosmo params from config.
    user_params = lib.UserParams(**cfg.get("user_params", {}))
    cosmo_params = lib.CosmoParams(**cfg.get("cosmo_params", {}))
    flag_options = lib.FlagOptions(**cfg.get("flag_options", {}))
    astro_params = lib.AstroParams(
        **cfg.get("astro_params", {}), INHOMO_RECO=flag_options.INHOMO_RECO
    )

    _override(ctx, user_params, cosmo_params, astro_params, flag_options)

    lib.run_lightcone(
        redshift=redshift,
        max_redshift=max_z,
        astro_params=astro_params,
        flag_options=flag_options,
        user_params=user_params,
        cosmo_params=cosmo_params,
        regenerate=regen,
        write=True,
        direc=direc,
        random_seed=seed,
    )


def _query(direc=None, kind=None, md5=None, seed=None, clear=False):
    cls = list(
        cache_tools.query_cache(direc=direc, kind=kind, hsh=md5, seed=seed, show=False)
    )

    if not clear:
        print("%s Data Sets Found:" % len(cls))
        print("------------------")
    else:
        print("Removing %s data sets..." % len(cls))

    for file, c in cls:
        if not clear:
            print("  @ {%s}:" % file)
            print("  %s" % str(c))

            print()

        else:
            direc = direc or path.expanduser(_cfg.config["direc"])
            remove(path.join(direc, file))


@main.command()
@click.option(
    "-d",
    "--direc",
    type=click.Path(exists=True, dir_okay=True),
    default=None,
    help="directory to write data and plots to -- must exist.",
)
@click.option("-k", "--kind", type=str, default=None, help="filter by kind of data.")
@click.option("-m", "--md5", type=str, default=None, help="filter by md5 hsh")
@click.option("-s", "--seed", type=str, default=None, help="filter by random seed")
@click.option(
    "--clear/--no-clear",
    default=False,
    help="remove all data sets returned by this query.",
)
def query(direc, kind, md5, seed, clear):
    """Query the cache database.

    Parameters
    ----------
    direc : str
        Directory in which to search for cache items
    kind : str
        Filter output by kind of box (eg. InitialConditions)
    md5 : str
        Filter output by hsh
    seed : float
        Filter output by random seed.
    clear : bool
        Remove all data sets returned by the query.
    """
    _query(direc, kind, md5, seed, clear)


@main.command()
@click.argument("param", type=str)
@click.argument("value", type=str)
@click.option(
    "-s",
    "--struct",
    type=click.Choice(["flag_options", "cosmo_params", "user_params", "astro_params"]),
    default="flag_options",
    help="struct in which the new feature exists",
)
@click.option(
    "-t",
    "--vtype",
    type=click.Choice(["bool", "float", "int"]),
    default="bool",
    help="type of the new parameter",
)
@click.option(
    "-l/-c",
    "--lightcone/--coeval",
    default=True,
    help="whether to use a lightcone for comparison",
)
@click.option(
    "-z", "--redshift", type=float, default=6.0, help="redshift of the comparison boxes"
)
@click.option(
    "-Z",
    "--max-redshift",
    type=float,
    default=30,
    help="maximum redshift of the comparison lightcone",
)
@click.option("-r", "--random-seed", type=int, default=12345, help="random seed to use")
@click.option("-v", "--verbose", count=True)
@click.option(
    "-g/-G",
    "--regenerate/--cache",
    default=True,
    help="whether to regenerate the boxes",
)
def pr_feature(
    param,
    value,
    struct,
    vtype,
    lightcone,
    redshift,
    max_redshift,
    random_seed,
    verbose,
    regenerate,
):
    """
    Create standard plots comparing a default simulation against a simulation with a new feature.

    The new feature is switched on by setting PARAM to VALUE.
    Plots are saved in the current directory, with the prefix "pr_feature".

    Parameters
    ----------
    param : str
        Name of the parameter to modify to "switch on" the feature.
    value : float
        Value to which to set it.
    struct : str
        The input parameter struct to which `param` belongs.
    vtype : str
        Type of the new parameter.
    lightcone : bool
        Whether the comparison should be done on a lightcone.
    redshift : float
        Redshift of comparison.
    max_redshift : float
        If using a lightcone, the maximum redshift in the lightcone to compare.
    random_seed : int
        Random seed at which to compare.
    verbose : int
        How verbose the output should be.
    regenerate : bool
        Whether to regenerate all data, even if it is in cache.
    """
    import powerbox

    lvl = [logging.WARNING, logging.INFO, logging.DEBUG][verbose]
    logger = logging.getLogger("21cmFAST")
    logger.setLevel(lvl)
    value = getattr(builtins, vtype)(value)

    structs = {
        "user_params": {"HII_DIM": 128, "BOX_LEN": 250},
        "flag_options": {"USE_TS_FLUCT": True},
        "cosmo_params": {},
        "astro_params": {},
    }

    if lightcone:
        print("Running default lightcone...")
        lc_default = lib.run_lightcone(
            redshift=redshift,
            max_redshift=max_redshift,
            random_seed=random_seed,
            regenerate=regenerate,
            **structs,
        )
        structs[struct][param] = value

        print("Running lightcone with new feature...")
        lc_new = lib.run_lightcone(
            redshift=redshift,
            max_redshift=max_redshift,
            random_seed=random_seed,
            regenerate=regenerate,
            **structs,
        )

        print("Plotting lightcone slices...")
        for field in ["brightness_temp"]:
            fig, ax = plt.subplots(3, 1, sharex=True, sharey=True)

            vmin = -150
            vmax = 30

            plotting.lightcone_sliceplot(
                lc_default, ax=ax[0], fig=fig, vmin=vmin, vmax=vmax
            )
            ax[0].set_title("Default")

            plotting.lightcone_sliceplot(
                lc_new, ax=ax[1], fig=fig, cbar=False, vmin=vmin, vmax=vmax
            )
            ax[1].set_title("New")

            plotting.lightcone_sliceplot(
                lc_default, lightcone2=lc_new, cmap="bwr", ax=ax[2], fig=fig
            )
            ax[2].set_title("Difference")

            plt.savefig("pr_feature_lighcone_2d_{}.pdf".format(field))

        def rms(x, axis=None):
            return np.sqrt(np.mean(x ** 2, axis=axis))

        print("Plotting lightcone history...")
        fig, ax = plt.subplots(4, 1, sharex=True, gridspec_kw={"hspace": 0.05})
        ax[0].plot(lc_default.node_redshifts, lc_default.global_xHI, label="Default")
        ax[0].plot(lc_new.node_redshifts, lc_new.global_xHI, label="New")
        ax[0].set_ylabel(r"$x_{\rm HI}$")
        ax[0].legend()

        ax[1].plot(
            lc_default.node_redshifts,
            lc_default.global_brightness_temp,
            label="Default",
        )
        ax[1].plot(lc_new.node_redshifts, lc_new.global_brightness_temp, label="New")
        ax[1].set_ylabel("$T_b$ [K]")
        ax[3].set_xlabel("z")

        rms_diff = rms(lc_default.brightness_temp, axis=(0, 1)) - rms(
            lc_new.brightness_temp, axis=(0, 1)
        )
        ax[2].plot(lc_default.lightcone_redshifts, rms_diff, label="RMS")
        ax[2].plot(
            lc_new.node_redshifts,
            lc_default.global_xHI - lc_new.global_xHI,
            label="$x_{HI}$",
        )
        ax[2].plot(
            lc_new.node_redshifts,
            lc_default.global_brightness_temp - lc_new.global_brightness_temp,
            label="$T_b$",
        )
        ax[2].legend()
        ax[2].set_ylabel("Differences")

        diff_rms = rms(lc_default.brightness_temp - lc_new.brightness_temp, axis=(0, 1))
        ax[3].plot(lc_default.lightcone_redshifts, diff_rms)
        ax[3].set_ylabel("RMS of Diff.")

        plt.savefig("pr_feature_history.pdf")

        print("Plotting power spectra history...")
        p_default = []
        p_new = []
        z = []
        thickness = 200  # Mpc
        ncells = int(thickness / lc_new.cell_size)
        chunk_size = lc_new.cell_size * ncells
        start = 0
        print(ncells)
        while start + ncells <= lc_new.shape[-1]:
            pd, k = powerbox.get_power(
                lc_default.brightness_temp[:, :, start : start + ncells],
                lc_default.lightcone_dimensions[:2] + (chunk_size,),
            )
            p_default.append(pd)

            pn, k = powerbox.get_power(
                lc_new.brightness_temp[:, :, start : start + ncells],
                lc_new.lightcone_dimensions[:2] + (chunk_size,),
            )
            p_new.append(pn)
            z.append(lc_new.lightcone_redshifts[start])

            start += ncells

        p_default = np.array(p_default).T
        p_new = np.array(p_new).T

        fig, ax = plt.subplots(2, 1, sharex=True)
        ax[0].set_yscale("log")

        inds = [
            np.where(np.abs(k - 0.1) == np.abs(k - 0.1).min())[0][0],
            np.where(np.abs(k - 0.2) == np.abs(k - 0.2).min())[0][0],
            np.where(np.abs(k - 0.5) == np.abs(k - 0.5).min())[0][0],
            np.where(np.abs(k - 1) == np.abs(k - 1).min())[0][0],
        ]

        for i, (pdef, pnew, kk) in enumerate(
            zip(p_default[inds], p_new[inds], k[inds])
        ):
            ax[0].plot(
                z, pdef, ls="--", label="k={:.2f}".format(kk), color="C{}".format(i)
            )
            ax[0].plot(z, pnew, ls="-", color="C{}".format(i))
            ax[1].plot(z, np.log10(pdef / pnew), ls="-", color="C{}".format(i))
        ax[1].set_xlabel("z")
        ax[0].set_ylabel(r"$\Delta^2 [{\rm mK}^2]$")
        ax[1].set_ylabel(r"log ratio of $\Delta^2 [{\rm mK}^2]$")
        ax[0].legend()

        plt.savefig("pr_feature_power_history.pdf")

    else:
        raise NotImplementedError()



"""
Output class objects.

The classes provided by this module exist to simplify access to large datasets created within C.
Fundamentally, ownership of the data belongs to these classes, and the C functions merely accesses
this and fills it. The various boxes and lightcones associated with each output are available as
instance attributes. Along with the output data, each output object contains the various input
parameter objects necessary to define it.

.. warning:: These should not be instantiated or filled by the user, but always handled
             as output objects from the various functions contained here. Only the data
             within the objects should be accessed.
"""
import os
import warnings
from hashlib import md5

import h5py
import numpy as np
from astropy import units
from astropy.cosmology import z_at_value
from cached_property import cached_property

from . import __version__
from . import _utils as _ut
from ._cfg import config
from ._utils import OutputStruct as _BaseOutputStruct
from ._utils import _check_compatible_inputs
from .c_21cmfast import ffi
from .c_21cmfast import lib
from .inputs import AstroParams
from .inputs import CosmoParams
from .inputs import FlagOptions
from .inputs import UserParams
from .inputs import global_params


class _OutputStruct(_BaseOutputStruct):
    _global_params = global_params

    def __init__(self, *, user_params=None, cosmo_params=None, **kwargs):
        if cosmo_params is None:
            cosmo_params = CosmoParams()
        if user_params is None:
            user_params = UserParams()

        super().__init__(user_params=user_params, cosmo_params=cosmo_params, **kwargs)

    _ffi = ffi


class _OutputStructZ(_OutputStruct):
    _inputs = _OutputStruct._inputs + ["redshift"]


class InitialConditions(_OutputStruct):
    """A class containing all initial conditions boxes."""

    _c_compute_function = lib.ComputeInitialConditions

    # The filter params indicates parameters to overlook when deciding if a cached box
    # matches current parameters.
    # It is useful for ignoring certain global parameters which may not apply to this
    # step or its dependents.
    _meta = False
    _filter_params = _OutputStruct._filter_params + [
        "ALPHA_UVB",  # ionization
        "EVOLVE_DENSITY_LINEARLY",  # perturb
        "SMOOTH_EVOLVED_DENSITY_FIELD",  # perturb
        "R_smooth_density",  # perturb
        "HII_ROUND_ERR",  # ionization
        "FIND_BUBBLE_ALGORITHM",  # ib
        "N_POISSON",  # ib
        "T_USE_VELOCITIES",  # bt
        "MAX_DVDR",  # bt
        "DELTA_R_HII_FACTOR",  # ib
        "HII_FILTER",  # ib
        "INITIAL_REDSHIFT",  # pf
        "HEAT_FILTER",  # st
        "CLUMPING_FACTOR",  # st
        "Z_HEAT_MAX",  # st
        "R_XLy_MAX",  # st
        "NUM_FILTER_STEPS_FOR_Ts",  # ts
        "ZPRIME_STEP_FACTOR",  # ts
        "TK_at_Z_HEAT_MAX",  # ts
        "XION_at_Z_HEAT_MAX",  # ts
        "Pop",  # ib
        "Pop2_ion",  # ib
        "Pop3_ion",  # ib
        "NU_X_BAND_MAX",  # st
        "NU_X_MAX",  # ib
    ]

    def _init_arrays(self):
        self.lowres_density = np.zeros(
            self.user_params.HII_tot_num_pixels, dtype=np.float32
        )
        self.lowres_vx = np.zeros(self.user_params.HII_tot_num_pixels, dtype=np.float32)
        self.lowres_vy = np.zeros(self.user_params.HII_tot_num_pixels, dtype=np.float32)
        self.lowres_vz = np.zeros(self.user_params.HII_tot_num_pixels, dtype=np.float32)

        self.hires_vx = np.zeros(self.user_params.tot_fft_num_pixels, dtype=np.float32)
        self.hires_vy = np.zeros(self.user_params.tot_fft_num_pixels, dtype=np.float32)
        self.hires_vz = np.zeros(self.user_params.tot_fft_num_pixels, dtype=np.float32)

        self.lowres_vx_2LPT = np.zeros(
            self.user_params.HII_tot_num_pixels, dtype=np.float32
        )
        self.lowres_vy_2LPT = np.zeros(
            self.user_params.HII_tot_num_pixels, dtype=np.float32
        )
        self.lowres_vz_2LPT = np.zeros(
            self.user_params.HII_tot_num_pixels, dtype=np.float32
        )

        self.hires_vx_2LPT = np.zeros(
            self.user_params.tot_fft_num_pixels, dtype=np.float32
        )
        self.hires_vy_2LPT = np.zeros(
            self.user_params.tot_fft_num_pixels, dtype=np.float32
        )
        self.hires_vz_2LPT = np.zeros(
            self.user_params.tot_fft_num_pixels, dtype=np.float32
        )

        self.hires_density = np.zeros(
            self.user_params.tot_fft_num_pixels, dtype=np.float32
        )

        self.hires_vcb = np.zeros(self.user_params.tot_fft_num_pixels, dtype=np.float32)
        self.lowres_vcb = np.zeros(
            self.user_params.HII_tot_num_pixels, dtype=np.float32
        )

        shape = (
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
        )
        hires_shape = (self.user_params.DIM, self.user_params.DIM, self.user_params.DIM)

        self.lowres_density.shape = shape
        self.lowres_vx.shape = shape
        self.lowres_vy.shape = shape
        self.lowres_vz.shape = shape
        self.lowres_vx_2LPT.shape = shape
        self.lowres_vy_2LPT.shape = shape
        self.lowres_vz_2LPT.shape = shape

        self.hires_density.shape = hires_shape
        self.hires_vx.shape = hires_shape
        self.hires_vy.shape = hires_shape
        self.hires_vz.shape = hires_shape
        self.hires_vx_2LPT.shape = hires_shape
        self.hires_vy_2LPT.shape = hires_shape
        self.hires_vz_2LPT.shape = hires_shape

        self.lowres_vcb.shape = shape
        self.hires_vcb.shape = hires_shape


class PerturbedField(_OutputStructZ):
    """A class containing all perturbed field boxes."""

    _c_compute_function = lib.ComputePerturbField

    _meta = False
    _filter_params = _OutputStruct._filter_params + [
        "ALPHA_UVB",  # ionization
        "HII_ROUND_ERR",  # ionization
        "FIND_BUBBLE_ALGORITHM",  # ib
        "N_POISSON",  # ib
        "T_USE_VELOCITIES",  # bt
        "MAX_DVDR",  # bt
        "DELTA_R_HII_FACTOR",  # ib
        "HII_FILTER",  # ib
        "HEAT_FILTER",  # st
        "CLUMPING_FACTOR",  # st
        "Z_HEAT_MAX",  # st
        "R_XLy_MAX",  # st
        "NUM_FILTER_STEPS_FOR_Ts",  # ts
        "ZPRIME_STEP_FACTOR",  # ts
        "TK_at_Z_HEAT_MAX",  # ts
        "XION_at_Z_HEAT_MAX",  # ts
        "Pop",  # ib
        "Pop2_ion",  # ib
        "Pop3_ion",  # ib
        "NU_X_BAND_MAX",  # st
        "NU_X_MAX",  # ib
    ]

    def _init_arrays(self):
        self.density = np.zeros(self.user_params.HII_tot_num_pixels, dtype=np.float32)
        self.velocity = np.zeros(self.user_params.HII_tot_num_pixels, dtype=np.float32)

        self.density.shape = (
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
        )
        self.velocity.shape = (
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
        )


class _AllParamsBox(_OutputStructZ):
    _meta = True
    _inputs = _OutputStructZ._inputs + ["flag_options", "astro_params"]

    _filter_params = _OutputStruct._filter_params + [
        "T_USE_VELOCITIES",  # bt
        "MAX_DVDR",  # bt
    ]

    def __init__(self, astro_params=None, flag_options=None, first_box=False, **kwargs):
        if flag_options is None:
            flag_options = FlagOptions()

        if astro_params is None:
            astro_params = AstroParams(INHOMO_RECO=flag_options.INHOMO_RECO)

        self.log10_Mturnover_ave = 0.0
        self.log10_Mturnover_MINI_ave = 0.0

        self.first_box = first_box
        if first_box:
            self.mean_f_coll = 0.0
            self.mean_f_coll_MINI = 0.0

        super().__init__(astro_params=astro_params, flag_options=flag_options, **kwargs)


class HaloField(_AllParamsBox):
    """A class containing all fields related to halos."""

    _c_based_pointers = (
        "halo_masses",
        "halo_coords",
        "mass_bins",
        "fgtrm",
        "sqrt_dfgtrm",
        "dndlm",
        "sqrtdn_dlm",
    )
    _c_compute_function = lib.ComputeHaloField
    _c_free_function = lib.free_halo_field

    def _init_arrays(self):
        self.halo_field = np.zeros(
            self.user_params.tot_fft_num_pixels, dtype=np.float32
        )

        self.halo_field.shape = (
            self.user_params.DIM,
            self.user_params.DIM,
            self.user_params.DIM,
        )

    def _c_shape(self, cstruct):
        return {
            "halo_masses": (cstruct.n_halos,),
            "halo_coords": (cstruct.n_halos, 3),
            "mass_bins": (cstruct.n_mass_bins,),
            "fgtrm": (cstruct.n_mass_bins,),
            "sqrt_dfgtrm": (cstruct.n_mass_bins,),
            "dndlm": (cstruct.n_mass_bins,),
            "sqrtdn_dlm": (cstruct.n_mass_bins,),
        }


class PerturbHaloField(_AllParamsBox):
    """A class containing all fields related to halos."""

    _c_compute_function = lib.ComputePerturbHaloField
    _c_based_pointers = ("halo_masses", "halo_coords")
    _c_free_function = lib.free_phf

    def _c_shape(self, cstruct):
        return {
            "halo_masses": (cstruct.n_halos,),
            "halo_coords": (cstruct.n_halos, 3),
        }


class IonizedBox(_AllParamsBox):
    """A class containing all ionized boxes."""

    _meta = False
    _c_compute_function = lib.ComputeIonizedBox

    def _init_arrays(self):
        if self.flag_options.USE_MINI_HALOS:
            Nfiltering = (
                int(
                    np.log(
                        min(
                            self.astro_params.R_BUBBLE_MAX,
                            0.620350491 * self.user_params.BOX_LEN,
                        )
                        / max(
                            global_params.R_BUBBLE_MIN,
                            0.620350491
                            * self.user_params.BOX_LEN
                            / self.user_params.HII_DIM,
                        )
                    )
                    / np.log(global_params.DELTA_R_HII_FACTOR)
                )
                + 1
            )
        else:
            Nfiltering = 1

        # ionized_box is always initialised to be neutral for excursion set algorithm.
        # Hence np.ones instead of np.zeros
        self.xH_box = np.ones(self.user_params.HII_tot_num_pixels, dtype=np.float32)
        self.Gamma12_box = np.zeros(
            self.user_params.HII_tot_num_pixels, dtype=np.float32
        )
        self.z_re_box = np.zeros(self.user_params.HII_tot_num_pixels, dtype=np.float32)
        self.dNrec_box = np.zeros(self.user_params.HII_tot_num_pixels, dtype=np.float32)
        self.temp_kinetic_all_gas = np.zeros(
            self.user_params.HII_tot_num_pixels, dtype=np.float32
        )
        self.Fcoll = np.zeros(
            Nfiltering * self.user_params.HII_tot_num_pixels, dtype=np.float32
        )

        shape = (
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
        )
        filter_shape = (
            Nfiltering,
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
        )
        self.xH_box.shape = shape
        self.Gamma12_box.shape = shape
        self.z_re_box.shape = shape
        self.dNrec_box.shape = shape
        self.temp_kinetic_all_gas.shape = shape
        self.Fcoll.shape = filter_shape

        if self.flag_options.USE_MINI_HALOS:
            self.Fcoll_MINI = np.zeros(
                Nfiltering * self.user_params.HII_tot_num_pixels, dtype=np.float32
            )
            self.Fcoll_MINI.shape = filter_shape
        else:
            self.Fcoll_MINI = np.array([], dtype=np.float32)

    @cached_property
    def global_xH(self):
        """Global (mean) neutral fraction."""
        if not self.filled:
            raise AttributeError(
                "global_xH is not defined until the ionization calculation has been performed"
            )
        else:
            return np.mean(self.xH_box)


class TsBox(_AllParamsBox):
    """A class containing all spin temperature boxes."""

    _c_compute_function = lib.ComputeTsBox
    _meta = False

    def _init_arrays(self):
        self.Ts_box = np.zeros(self.user_params.HII_tot_num_pixels, dtype=np.float32)
        self.x_e_box = np.zeros(self.user_params.HII_tot_num_pixels, dtype=np.float32)
        self.Tk_box = np.zeros(self.user_params.HII_tot_num_pixels, dtype=np.float32)
        self.J_21_LW_box = np.zeros(
            self.user_params.HII_tot_num_pixels, dtype=np.float32
        )
        shape = (
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
        )

        self.Ts_box.shape = shape
        self.x_e_box.shape = shape
        self.Tk_box.shape = shape
        self.J_21_LW_box.shape = shape

    @cached_property
    def global_Ts(self):
        """Global (mean) spin temperature."""
        if not self.filled:
            raise AttributeError(
                "global_Ts is not defined until the ionization calculation has been performed"
            )
        else:
            return np.mean(self.Ts_box)

    @cached_property
    def global_Tk(self):
        """Global (mean) Tk."""
        if not self.filled:
            raise AttributeError(
                "global_Tk is not defined until the ionization calculation has been performed"
            )
        else:
            return np.mean(self.Tk_box)

    @cached_property
    def global_x_e(self):
        """Global (mean) x_e."""
        if not self.filled:
            raise AttributeError(
                "global_x_e is not defined until the ionization calculation has been performed"
            )
        else:
            return np.mean(self.x_e_box)


class BrightnessTemp(_AllParamsBox):
    """A class containing the brightness temperature box."""

    _c_compute_function = lib.ComputeBrightnessTemp

    _meta = False
    _filter_params = _OutputStructZ._filter_params

    def _init_arrays(self):
        self.brightness_temp = np.zeros(
            self.user_params.HII_tot_num_pixels, dtype=np.float32
        )

        self.brightness_temp.shape = (
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
            self.user_params.HII_DIM,
        )

    @cached_property
    def global_Tb(self):
        """Global (mean) brightness temperature."""
        if not self.filled:
            raise AttributeError(
                "global_Tb is not defined until the ionization calculation has been performed"
            )
        else:
            return np.mean(self.brightness_temp)


class _HighLevelOutput:
    def _get_prefix(self):
        return "{name}_z{redshift:.4}_{{hash}}_r{seed}.h5".format(
            name=self.__class__.__name__,
            redshift=float(self.redshift),
            seed=self.random_seed,
        )

    def _input_rep(self):
        rep = ""
        for inp in [
            "user_params",
            "cosmo_params",
            "astro_params",
            "flag_options",
            "global_params",
        ]:
            rep += repr(getattr(self, inp))
        return rep

    def get_unique_filename(self):
        """Generate a unique hash filename for this instance."""
        return self._get_prefix().format(
            hash=md5((self._input_rep() + self._particular_rep()).encode()).hexdigest()
        )

    def _write(self, direc=None, fname=None, clobber=False):
        """
        Write the lightcone to file in standard HDF5 format.

        This method is primarily meant for the automatic caching. Its default
        filename is a hash generated based on the input data, and the directory is
        the configured caching directory.

        Parameters
        ----------
        direc : str, optional
            The directory into which to write the file. Default is the configuration
            directory.
        fname : str, optional
            The filename to write, default a unique name produced by the inputs.
        clobber : bool, optional
            Whether to overwrite existing file.

        Returns
        -------
        fname : str
            The absolute path to which the file was written.
        """
        direc = os.path.expanduser(direc or config["direc"])

        if fname is None:
            fname = self.get_unique_filename()

        if not os.path.isabs(fname):
            fname = os.path.abspath(os.path.join(direc, fname))

        if not clobber and os.path.exists(fname):
            raise FileExistsError(
                "The file {} already exists. If you want to overwrite, set clobber=True.".format(
                    fname
                )
            )

        with h5py.File(fname, "w") as f:
            # Save input parameters as attributes
            for k in [
                "user_params",
                "cosmo_params",
                "flag_options",
                "astro_params",
                "global_params",
            ]:
                q = getattr(self, k)
                kfile = "_globals" if k == "global_params" else k
                grp = f.create_group(kfile)

                try:
                    dct = q.self
                except AttributeError:
                    dct = q

                for kk, v in dct.items():
                    if v is None:
                        continue
                    try:
                        grp.attrs[kk] = v
                    except TypeError:
                        # external_table_path is a cdata object and can't be written.
                        pass

            if self.photon_nonconservation_data is not None:
                photon_data = f.create_group("photon_nonconservation_data")
                for k, val in self.photon_nonconservation_data.items():
                    photon_data[k] = val

            f.attrs["redshift"] = self.redshift
            f.attrs["random_seed"] = self.random_seed
            f.attrs["version"] = __version__

        self._write_particulars(fname)

        return fname

    def _write_particulars(self, fname):
        pass

    def save(self, fname=None, direc="."):
        """Save to disk.

        This function has defaults that make it easy to save a unique box to
        the current directory.

        Parameters
        ----------
        fname : str, optional
            The filename to write, default a unique name produced by the inputs.
        direc : str, optional
            The directory into which to write the file. Default is the current directory.

        Returns
        -------
        str :
            The filename to which the box was written.
        """
        return self._write(direc=direc, fname=fname)

    @classmethod
    def _read_inputs(cls, fname):
        kwargs = {}
        with h5py.File(fname, "r") as fl:
            glbls = dict(fl["_globals"].attrs)
            kwargs["redshift"] = fl.attrs["redshift"]

            if "photon_nonconservation_data" in fl.keys():
                d = fl["photon_nonconservation_data"]
                kwargs["photon_nonconservation_data"] = {k: d[k][...] for k in d.keys()}

        return kwargs, glbls

    @classmethod
    def read(cls, fname, direc="."):
        """Read a lightcone file from disk, creating a LightCone object.

        Parameters
        ----------
        fname : str
            The filename path. Can be absolute or relative.
        direc : str
            If fname, is relative, the directory in which to find the file. By default,
            both the current directory and default cache and the  will be searched, in
            that order.

        Returns
        -------
        LightCone :
            A :class:`LightCone` instance created from the file's data.
        """
        if not os.path.isabs(fname):
            fname = os.path.abspath(os.path.join(direc, fname))

        if not os.path.exists(fname):
            raise FileExistsError("The file {} does not exist!".format(fname))

        park, glbls = cls._read_inputs(fname)
        boxk = cls._read_particular(fname)

        with global_params.use(**glbls):
            out = cls(**park, **boxk)

        return out


class Coeval(_HighLevelOutput):
    """A full coeval box with all associated data."""

    def __init__(
        self,
        redshift: float,
        initial_conditions: InitialConditions,
        perturbed_field: PerturbedField,
        ionized_box: IonizedBox,
        brightness_temp: BrightnessTemp,
        ts_box: [TsBox, None] = None,
        photon_nonconservation_data=None,
        _globals=None,
    ):
        _check_compatible_inputs(
            initial_conditions,
            perturbed_field,
            ionized_box,
            brightness_temp,
            ts_box,
            ignore=[],
        )

        self.redshift = redshift
        self.init_struct = initial_conditions
        self.perturb_struct = perturbed_field
        self.ionization_struct = ionized_box
        self.brightness_temp_struct = brightness_temp
        self.spin_temp_struct = ts_box

        self.photon_nonconservation_data = photon_nonconservation_data
        # A *copy* of the current global parameters.
        self.global_params = _globals or dict(global_params.items())

        # Expose all the fields of the structs to the surface of the Coeval object
        for box in [
            initial_conditions,
            perturbed_field,
            ionized_box,
            brightness_temp,
            ts_box,
        ]:
            if box is None:
                continue
            for field in box.fieldnames:
                setattr(self, field, getattr(box, field))

    @property
    def user_params(self):
        """User params shared by all datasets."""
        return self.brightness_temp_struct.user_params

    @property
    def cosmo_params(self):
        """Cosmo params shared by all datasets."""
        return self.brightness_temp_struct.cosmo_params

    @property
    def flag_options(self):
        """Flag Options shared by all datasets."""
        return self.brightness_temp_struct.flag_options

    @property
    def astro_params(self):
        """Astro params shared by all datasets."""
        return self.brightness_temp_struct.astro_params

    @property
    def random_seed(self):
        """Random seed shared by all datasets."""
        return self.brightness_temp_struct.random_seed

    def _particular_rep(self):
        return ""

    def _write_particulars(self, fname):
        for name in ["init", "perturb", "ionization", "brightness_temp", "spin_temp"]:
            struct = getattr(self, name + "_struct")
            if struct is not None:
                struct.write(fname=fname, write_inputs=False)

    @classmethod
    def _read_particular(cls, fname):
        kwargs = {}

        with h5py.File(fname, "r") as fl:
            for output_class in _ut.OutputStruct._implementations():
                if output_class.__name__ in fl:
                    kwargs[
                        _ut.camel_to_snake(output_class.__name__)
                    ] = output_class.from_file(fname)

        return kwargs

    def __eq__(self, other):
        """Determine if this is equal to another object."""
        return (
            isinstance(other, self.__class__)
            and other.redshift == self.redshift
            and self.user_params == other.user_params
            and self.cosmo_params == other.cosmo_params
            and self.flag_options == other.flag_options
            and self.astro_params == other.astro_params
        )


class LightCone(_HighLevelOutput):
    """A full Lightcone with all associated evolved data."""

    def __init__(
        self,
        redshift,
        user_params,
        cosmo_params,
        astro_params,
        flag_options,
        random_seed,
        lightcones,
        node_redshifts=None,
        global_quantities=None,
        photon_nonconservation_data=None,
        _globals=None,
    ):
        self.redshift = redshift
        self.random_seed = random_seed
        self.user_params = user_params
        self.cosmo_params = cosmo_params
        self.astro_params = astro_params
        self.flag_options = flag_options
        self.node_redshifts = node_redshifts

        # A *copy* of the current global parameters.
        self.global_params = _globals or dict(global_params.items())

        if global_quantities:
            for name, data in global_quantities.items():
                if name.endswith("_box"):
                    # Remove the _box because it looks dumb.
                    setattr(self, "global_" + name[:-4], data)
                else:
                    setattr(self, "global_" + name, data)

        self.photon_nonconservation_data = photon_nonconservation_data

        for name, data in lightcones.items():
            setattr(self, name, data)

        # Hold a reference to the global/lightcones in a dict form for easy reference.
        self.global_quantities = global_quantities
        self.lightcones = lightcones

    @property
    def global_xHI(self):
        """Global neutral fraction function."""
        warnings.warn(
            "global_xHI is deprecated. From now on, use global_xH. Will be removed in v3.1"
        )
        return self.global_xH

    @property
    def cell_size(self):
        """Cell size [Mpc] of the lightcone voxels."""
        return self.user_params.BOX_LEN / self.user_params.HII_DIM

    @property
    def lightcone_dimensions(self):
        """Lightcone size over each dimension -- tuple of (x,y,z) in Mpc."""
        return (
            self.user_params.BOX_LEN,
            self.user_params.BOX_LEN,
            self.n_slices * self.cell_size,
        )

    @property
    def shape(self):
        """Shape of the lightcone as a 3-tuple."""
        return self.brightness_temp.shape

    @property
    def n_slices(self):
        """Number of redshift slices in the lightcone."""
        return self.shape[-1]

    @property
    def lightcone_coords(self):
        """Co-ordinates [Mpc] of each cell along the redshift axis."""
        return np.linspace(0, self.lightcone_dimensions[-1], self.n_slices)

    @property
    def lightcone_distances(self):
        """Comoving distance to each cell along the redshift axis, from z=0."""
        return (
            self.cosmo_params.cosmo.comoving_distance(self.redshift).value
            + self.lightcone_coords
        )

    @property
    def lightcone_redshifts(self):
        """Redshift of each cell along the redshift axis."""
        return np.array(
            [
                z_at_value(self.cosmo_params.cosmo.comoving_distance, d * units.Mpc)
                for d in self.lightcone_distances
            ]
        )

    def _particular_rep(self):
        return (
            str(np.round(self.node_redshifts, 3))
            + str(self.global_quantities.keys())
            + str(self.lightcones.keys())
        )

    def _write_particulars(self, fname):
        with h5py.File(fname, "a") as f:
            # Save the boxes to the file
            boxes = f.create_group("lightcones")

            # Go through all fields in this struct, and save
            for k, val in self.lightcones.items():
                boxes[k] = val

            global_q = f.create_group("global_quantities")
            for k, v in self.global_quantities.items():
                global_q[k] = v

            f["node_redshifts"] = self.node_redshifts

    @classmethod
    def _read_inputs(cls, fname):
        kwargs = {}
        with h5py.File(fname, "r") as fl:
            for (k, kls) in [
                ("user_params", UserParams),
                ("cosmo_params", CosmoParams),
                ("flag_options", FlagOptions),
                ("astro_params", AstroParams),
            ]:
                grp = fl[k]
                kwargs[k] = kls(dict(grp.attrs))
            kwargs["random_seed"] = fl.attrs["random_seed"]

        # Get the standard inputs.
        kw, glbls = _HighLevelOutput._read_inputs(fname)
        return {**kw, **kwargs}, glbls

    @classmethod
    def _read_particular(cls, fname):
        kwargs = {}
        with h5py.File(fname, "r") as fl:
            boxes = fl["lightcones"]
            kwargs["lightcones"] = {k: boxes[k][...] for k in boxes.keys()}

            glb = fl["global_quantities"]
            kwargs["global_quantities"] = {k: glb[k][...] for k in glb.keys()}

            kwargs["node_redshifts"] = fl["node_redshifts"][...]

        return kwargs

    def __eq__(self, other):
        """Determine if this is equal to another object."""
        return (
            isinstance(other, self.__class__)
            and other.redshift == self.redshift
            and np.all(np.isclose(other.node_redshifts, self.node_redshifts, atol=1e-3))
            and self.user_params == other.user_params
            and self.cosmo_params == other.cosmo_params
            and self.flag_options == other.flag_options
            and self.astro_params == other.astro_params
            and self.global_quantities.keys() == other.global_quantities.keys()
            and self.lightcones.keys() == other.lightcones.keys()
        )



"""Open and read the configuration file."""
import contextlib
import copy
import warnings
from os import path

from . import yaml


class ConfigurationError(Exception):
    """An error with the config file."""

    pass


class Config(dict):
    """Simple over-ride of dict that adds a context manager."""

    _defaults = {"direc": "~/21cmFAST-cache", "regenerate": False, "write": True}

    _aliases = {"direc": ("boxdir",)}

    def __init__(self, *args, write=True, **kwargs):

        super().__init__(*args, **kwargs)

        # Ensure the keys that got read in are the right keys for the current version
        do_write = False
        for k, v in self._defaults.items():
            if k not in self:
                if k not in self._aliases:
                    warnings.warn("Your configuration file is out of date. Updating...")
                    do_write = True
                    self[k] = v

                else:
                    for alias in self._aliases[k]:
                        if alias in self:
                            do_write = True
                            warnings.warn(
                                "Your configuration file has old key '{}' which has been re-named '{}'. Updating...".format(
                                    alias, k
                                )
                            )
                            self[k] = self[alias]
                            del self[alias]
                            break
                    else:
                        warnings.warn(
                            "Your configuration file is out of date. Updating..."
                        )
                        do_write = True
                        self[k] = v

        for k, v in self.items():
            if k not in self._defaults:
                raise ConfigurationError(
                    f"The configuration file has key '{alias}' which is not known to 21cmFAST."
                )

        if do_write and write:
            self.write()

    @contextlib.contextmanager
    def use(self, **kwargs):
        """Context manager for using certain configuration options for a set time."""
        backup = self.copy()
        for k, v in kwargs.items():
            self[k] = v
        yield self
        for k in kwargs:
            self[k] = backup[k]

    def write(self, fname=None):
        """Write current configuration to file to make it permanent."""
        fname = fname or self.file_name
        with open(fname, "w") as fl:
            yaml.dump(self._as_dict(), fl)

    def _as_dict(self):
        """The plain dict defining the instance."""
        return {k: v for k, v in self.items()}

    @classmethod
    def load(cls, file_name):
        """Create a Config object from a config file."""
        cls.file_name = file_name
        if path.exists(file_name):
            with open(file_name, "r") as fl:
                config = yaml.load(fl)
                return cls(config)
        else:
            return cls(write=False)


config = Config.load(path.expanduser(path.join("~", ".21cmfast", "config.yml")))

# Keep an original copy around
default_config = copy.deepcopy(config)



"""Simple plotting functions for 21cmFAST objects."""
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
from astropy import units as un
from astropy.cosmology import z_at_value
from matplotlib import colors
from matplotlib.ticker import AutoLocator

from . import outputs
from .outputs import Coeval
from .outputs import LightCone

eor_colour = colors.LinearSegmentedColormap.from_list(
    "EoR",
    [
        (0, "white"),
        (0.21, "yellow"),
        (0.42, "orange"),
        (0.63, "red"),
        (0.86, "black"),
        (0.9, "blue"),
        (1, "cyan"),
    ],
)
plt.register_cmap(cmap=eor_colour)


def _imshow_slice(
    cube,
    slice_axis=-1,
    slice_index=0,
    fig=None,
    ax=None,
    fig_kw=None,
    cbar=True,
    cbar_horizontal=False,
    rotate=False,
    cmap="EoR",
    log: [bool] = False,
    **imshow_kw,
):
    """
    Plot a slice of some kind of cube.

    Parameters
    ----------
    cube : nd-array
        A 3D array of some quantity.
    slice_axis : int, optional
        The axis over which to take a slice, in order to plot.
    slice_index :
        The index of the slice.
    fig : Figure object
        An optional matplotlib figure object on which to plot
    ax : Axis object
        The matplotlib axis object on which to plot (created by default).
    fig_kw :
        Optional arguments passed to the figure construction.
    cbar : bool
        Whether to plot the colorbar
    cbar_horizontal : bool
        Whether the colorbar should be horizontal underneath the plot.
    rotate : bool
        Whether to rotate the plot vertically.
    imshow_kw :
        Optional keywords to pass to :func:`maplotlib.imshow`.

    Returns
    -------
    fig, ax :
        The figure and axis objects from matplotlib.
    """
    # If no axis is passed, create a new one
    # This allows the user to add this plot into an existing grid, or alter it afterwards.
    if fig_kw is None:
        fig_kw = {}
    if ax is None and fig is None:
        fig, ax = plt.subplots(1, 1, **fig_kw)
    elif ax is None:
        ax = plt.gca()
    elif fig is None:
        fig = plt.gcf()

    plt.sca(ax)

    if slice_index >= cube.shape[slice_axis]:
        raise IndexError(
            "slice_index is too large for that axis (slice_index=%s >= %s"
            % (slice_index, cube.shape[slice_axis])
        )

    slc = np.take(cube, slice_index, axis=slice_axis)
    if not rotate:
        slc = slc.T

    if cmap == "EoR":
        imshow_kw["vmin"] = -150
        imshow_kw["vmax"] = 30

    norm = imshow_kw.get("norm", colors.LogNorm() if log else colors.Normalize())
    plt.imshow(slc, origin="lower", cmap=cmap, norm=norm, **imshow_kw)

    if cbar:
        cb = plt.colorbar(
            orientation="horizontal" if cbar_horizontal else "vertical", aspect=40
        )
        cb.outline.set_edgecolor(None)

    return fig, ax


def coeval_sliceplot(
    struct: [outputs._OutputStruct, Coeval],
    kind: [str, None] = None,
    cbar_label: [str, None] = None,
    **kwargs,
):
    """
    Show a slice of a given coeval box.

    Parameters
    ----------
    struct : :class:`~outputs._OutputStruct` or :class:`~wrapper.Coeval` instance
        The output of a function such as `ionize_box` (a class containing several quantities), or
        `run_coeval`.
    kind : str
        The quantity within the structure to be shown.
    cbar_label : str, optional
        A label for the colorbar. Some values of `kind` will have automatically chosen
        labels, but these can be turned off by setting ``cbar_label=''``.

    Returns
    -------
    fig, ax :
        figure and axis objects from matplotlib

    Other Parameters
    ----------------
    All other parameters are passed directly to :func:`_imshow_slice`. These include `slice_axis`
    and `slice_index`,
    which choose the actual slice to plot, optional `fig` and `ax` keywords which enable
    over-plotting previous figures,
    and the `imshow_kw` argument, which allows arbitrary styling of the plot.
    """
    if kind is None:
        if isinstance(struct, outputs._OutputStruct):
            kind = struct.fieldnames[0]
        elif isinstance(struct, Coeval):
            kind = "brightness_temp"

    try:
        cube = getattr(struct, kind)
    except AttributeError:
        raise AttributeError(
            "The given OutputStruct does not have the quantity {kind}".format(kind=kind)
        )

    if kind != "brightness_temp" and "cmap" not in kwargs:
        kwargs["cmap"] = "viridis"

    fig, ax = _imshow_slice(cube, extent=(0, struct.user_params.BOX_LEN) * 2, **kwargs)

    slice_axis = kwargs.get("slice_axis", -1)

    # Determine which axes are being plotted.
    if slice_axis in (2, -1):
        xax = "x"
        yax = "y"
    elif slice_axis == 1:
        xax = "x"
        yax = "z"
    elif slice_axis == 0:
        xax = "y"
        yax = "z"
    else:
        raise ValueError("slice_axis should be between -1 and 2")

    # Now put on the decorations.
    ax.set_xlabel("{xax}-axis [Mpc]".format(xax=xax))
    ax.set_ylabel("{yax}-axis [Mpc]".format(yax=yax))

    cbar = fig._gci().colorbar

    if cbar is not None:
        if cbar_label is None:
            if kind == "brightness_temp":
                cbar_label = r"Brightness Temperature, $\delta T_B$ [mK]"
            elif kind == "xH_box":
                cbar_label = r"Neutral fraction"

        cbar.ax.set_ylabel(cbar_label)

    return fig, ax


def lightcone_sliceplot(
    lightcone: LightCone,
    kind: str = "brightness_temp",
    lightcone2: LightCone = None,
    vertical: bool = False,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    cbar_label: Optional[str] = None,
    zticks: str = "redshift",
    fig: Optional[plt.Figure] = None,
    ax: Optional[plt.Axes] = None,
    **kwargs,
):
    """Create a 2D plot of a slice through a lightcone.

    Parameters
    ----------
    lightcone : :class:`~py21cmfast.wrapper.Lightcone`
        The lightcone object to plot
    kind : str, optional
        The attribute of the lightcone to plot. Must be an array.
    lightcone2 : str, optional
        If provided, plot the _difference_ of the selected attribute between the two
        lightcones.
    vertical : bool, optional
        Whether to plot the redshift in the vertical direction.
    cbar_label : str, optional
        A label for the colorbar. Some quantities have automatically chosen labels, but
        these can be removed by setting `cbar_label=''`.
    zticks : str, optional
        Defines the co-ordinates of the ticks along the redshift axis.
        Can be "redshift" (default), "frequency", "distance" (which starts at zero
        for the lowest redshift) or the name of any function in an astropy cosmology
        that is purely a function of redshift.
    kwargs :
        Passed through to ``imshow()``.

    Returns
    -------
    fig :
        The matplotlib Figure object
    ax :
        The matplotlib Axis object onto which the plot was drawn.
    """
    slice_axis = kwargs.pop("slice_axis", 0)
    if slice_axis <= -2 or slice_axis >= 3:
        raise ValueError(
            "slice_axis should be between -1 and 2 (got {})".format(slice_axis)
        )

    z_axis = ("y" if vertical else "x") if slice_axis in (0, 1) else None

    # Dictionary mapping axis to dimension in lightcone
    axis_dct = {
        "x": 2 if z_axis == "x" else [1, 0, 0][slice_axis],
        "y": 2 if z_axis == "y" else [1, 0, 1][slice_axis],
    }

    if fig is None and ax is None:
        fig, ax = plt.subplots(
            1,
            1,
            figsize=(
                lightcone.shape[axis_dct["x"]] * 0.015 + 0.5,
                lightcone.shape[axis_dct["y"]] * 0.015
                + (2.5 if kwargs.get("cbar", True) else 0.05),
            ),
        )
    elif fig is None:
        fig = ax._gci().figure
    elif ax is None:
        ax = fig.get_axes()

    # Get x,y labels if they're not the redshift axis.
    if xlabel is None:
        xlabel = (
            None if axis_dct["x"] == 2 else "{}-axis [Mpc]".format("xy"[axis_dct["x"]])
        )
    if ylabel is None:
        ylabel = (
            None if axis_dct["y"] == 2 else "{}-axis [Mpc]".format("xy"[axis_dct["y"]])
        )

    extent = (
        0,
        lightcone.lightcone_dimensions[axis_dct["x"]],
        0,
        lightcone.lightcone_dimensions[axis_dct["y"]],
    )

    if lightcone2 is None:
        fig, ax = _imshow_slice(
            getattr(lightcone, kind),
            extent=extent,
            slice_axis=slice_axis,
            rotate=not vertical,
            cbar_horizontal=not vertical,
            cmap=kwargs.get("cmap", "EoR" if kind == "brightness_temp" else "viridis"),
            fig=fig,
            ax=ax,
            **kwargs,
        )
    else:
        d = getattr(lightcone, kind) - getattr(lightcone2, kind)
        fig, ax = _imshow_slice(
            d,
            extent=extent,
            slice_axis=slice_axis,
            rotate=not vertical,
            cbar_horizontal=not vertical,
            cmap=kwargs.pop("cmap", "bwr"),
            vmin=-np.abs(d.max()),
            vmax=np.abs(d.max()),
            fig=fig,
            ax=ax,
            **kwargs,
        )

    if z_axis:
        zlabel = _set_zaxis_ticks(ax, lightcone, zticks, z_axis)

    if ylabel != "":
        ax.set_ylabel(ylabel or zlabel)
    if xlabel != "":
        ax.set_xlabel(xlabel or zlabel)

    cbar = fig._gci().colorbar

    if cbar_label is None:
        if kind == "brightness_temp":
            cbar_label = r"Brightness Temperature, $\delta T_B$ [mK]"
        elif kind == "xH":
            cbar_label = r"Neutral fraction"

    if vertical:
        cbar.ax.set_ylabel(cbar_label)
    else:
        cbar.ax.set_xlabel(cbar_label)

    return fig, ax


def _set_zaxis_ticks(ax, lightcone, zticks, z_axis):
    if zticks != "distance":
        loc = AutoLocator()
        # Get redshift ticks.
        lc_z = lightcone.lightcone_redshifts

        if zticks == "redshift":
            coords = lc_z
        elif zticks == "frequency":
            coords = 1420 / (1 + lc_z) * un.MHz
        else:
            try:
                coords = getattr(lightcone.cosmo_params.cosmo, zticks)(lc_z)
            except AttributeError:
                raise AttributeError(
                    "zticks '{}' is not a cosmology function.".format(zticks)
                )

        zlabel = " ".join(z.capitalize() for z in zticks.split("_"))
        units = getattr(coords, "unit", None)
        if units:
            zlabel += " [{}]".format(str(coords.unit))
            coords = coords.value

        ticks = loc.tick_values(coords.min(), coords.max())

        if ticks.min() < coords.min() / 1.00001:
            ticks = ticks[1:]
        if ticks.max() > coords.max() * 1.00001:
            ticks = ticks[:-1]

        if coords[1] < coords[0]:
            ticks = ticks[::-1]

        if zticks == "redshift":
            z_ticks = ticks
        elif zticks == "frequency":
            z_ticks = 1420 / ticks - 1
        else:
            z_ticks = [
                z_at_value(getattr(lightcone.cosmo_params.cosmo, zticks), z * units)
                for z in ticks
            ]

        d_ticks = (
            lightcone.cosmo_params.cosmo.comoving_distance(z_ticks).value
            - lightcone.lightcone_distances[0]
        )
        getattr(ax, "set_{}ticks".format(z_axis))(d_ticks)
        getattr(ax, "set_{}ticklabels".format(z_axis))(ticks)

    else:
        zlabel = "Line-of-Sight Distance [Mpc]"
    return zlabel


def plot_global_history(
    lightcone: [LightCone],
    kind: [str, None] = None,
    ylabel: [str, None] = None,
    ylog: [bool] = False,
    ax: [plt.Axes, None] = None,
):
    """
    Plot the global history of a given quantity from a lightcone.

    Parameters
    ----------
    lightcone : :class:`~LightCone` instance
        The lightcone containing the quantity to plot.
    kind : str, optional
        The quantity to plot. Must be in the `global_quantities` dict in the lightcone.
        By default, will choose the first entry in the dict.
    ylabel : str, optional
        A y-label for the plot. If None, will use ``kind``.
    ax : Axes, optional
        The matplotlib Axes object on which to plot. Otherwise, created.
    """
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(4, 7))
    else:
        fig = ax._gci().figure

    if kind is None:
        kind = list(lightcone.global_quantities.keys())[0]

    assert (
        kind in lightcone.global_quantities
        or hasattr(lightcone, "global_" + kind)
        or (kind.startswith("global_") and hasattr(lightcone, kind))
    )

    if kind in lightcone.global_quantities:
        value = lightcone.global_quantities[kind]
    elif kind.startswith("global)"):
        value = getattr(lightcone, kind)
    else:
        value = getattr(lightcone, "global_" + kind)

    ax.plot(lightcone.node_redshifts, value)
    ax.set_xlabel("Redshift")
    if ylabel is None:
        ylabel = kind
    if ylabel:
        ax.set_ylabel(ylabel)

    if ylog:
        ax.set_yscale("log")

    return fig, ax
